# BSD 3-Clause License
# Copyright (c) 2024, mac
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

import os


class CrypterConfig():
    CONFIG_DIR = os.environ['HOME']
    CRYPTER_DIR = ".crypter"
    DB_NAME = "crypter"


# BSD 3-Clause License
# Copyright (c) 2024, mac
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

import os
import json
import random
import logging
from src.crypter_db import CrypterDb
from prettytable import PrettyTable
from src.crypter_config import CrypterConfig
from src.crypter_token_generator import CrypterTokenGenerator


class CrypterMain():
    @classmethod
    def init(cls):
        config_dir = os.path.join(CrypterConfig.CONFIG_DIR, CrypterConfig.CRYPTER_DIR)
        os.makedirs(config_dir, exist_ok=True)
        with CrypterDb(db_name=CrypterConfig.DB_NAME, db_path=config_dir) as db:
            db.setup()
            response = db.insert("secret", {'key': CrypterTokenGenerator.generate_key()}, return_columns=['Id'])
            return cls.format_response(response, return_columns=['Id'])

    @classmethod
    def add_key(cls, key_name, user_name, user_password, output_format='json'):
        config_dir = os.path.join(CrypterConfig.CONFIG_DIR, CrypterConfig.CRYPTER_DIR)
        with CrypterDb(db_name=CrypterConfig.DB_NAME, db_path=config_dir) as db:
            salt = list(key_name + user_name)
            random.shuffle(salt)
            user_password = user_password if user_password else CrypterTokenGenerator.generate_password(''.join(salt))
            values = {'key': key_name, 'name': user_name, 'password': user_password}
            return_columns = ['key', 'name', 'password']
            response = db.insert(table_name='records', values=values, return_columns=return_columns)
            return cls.format_response(response, return_columns=return_columns, output_format=output_format)

    @classmethod
    def get_key(cls, table_name='records', key_names=None, output_format='json'):
        config_dir = os.path.join(CrypterConfig.CONFIG_DIR, CrypterConfig.CRYPTER_DIR)
        with CrypterDb(db_name=CrypterConfig.DB_NAME, db_path=config_dir) as db:
            return_columns = ['key', 'name', 'password']
            response = db.get(table_name=table_name, key_names=key_names, return_columns=return_columns)
            return cls.format_response(response, return_columns=return_columns, output_format=output_format)

    @classmethod
    def delete_key(cls, table_name='records', key_names=None, output_format='json'):
        config_dir = os.path.join(CrypterConfig.CONFIG_DIR, CrypterConfig.CRYPTER_DIR)
        with CrypterDb(db_name=CrypterConfig.DB_NAME, db_path=config_dir) as db:
            return_columns = ['key', 'name', 'password']
            response = db.delete(table_name=table_name, key_names=key_names, return_columns=return_columns)
            return cls.format_response(response, return_columns=return_columns, output_format=output_format)

    @classmethod
    def _json_format_response(cls, query_result, return_columns):
        query_response = list()
        for row in query_result:
            r = dict()
            for col_name, col_value in zip(return_columns, list(row)):
                r.update({col_name: col_value})
            query_response.append(r)
        return json.dumps(query_response, indent=4)

    @classmethod
    def _tabular_format_response(cls, query_result, return_columns):
        table_data = PrettyTable(return_columns)
        for row in query_result:
            table_data.add_row(row)
        return table_data

    @classmethod
    def format_response(cls, query_result, return_columns, output_format='json'):
        if output_format == 'json':
            return cls._json_format_response(query_result, return_columns)

        if output_format == 'tabular':
            return cls._tabular_format_response(query_result, return_columns)

    @classmethod
    def cloud_init(cls):
        pass

    @classmethod
    def cloud_sync(cls):
        pass


# BSD 3-Clause License
# Copyright (c) 2024, mac
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

import re
import base64
import random
import string
import hashlib
import secrets
from cryptography.fernet import Fernet


class CrypterTokenGenerator():
    @classmethod
    def generate_key(cls):
        return Fernet.generate_key()

    @classmethod
    def generate_password(cls, salt, chunk_size=4):
        sha256Hash = hashlib.sha256()
        sha256Hash.update(salt.encode())
        sha256HexDigest = sha256Hash.hexdigest()
        passwordString = base64.urlsafe_b64encode(sha256HexDigest.encode())
        lower_case_in_password = re.findall(f"[{string.ascii_lowercase}]", passwordString.decode())
        upper_case_in_password = re.findall(f"[{string.ascii_uppercase}]", passwordString.decode())
        digits_in_password = re.findall(f"[{string.digits}]", passwordString.decode())
        special_chars_in_password = re.findall(f"[{string.punctuation}]", passwordString.decode())
        special_chars_in_password.extend([secrets.choice(string.punctuation) for i in range(chunk_size)])
        random.shuffle(lower_case_in_password);random.shuffle(upper_case_in_password)
        random.shuffle(digits_in_password);random.shuffle(special_chars_in_password)
        final_password = lower_case_in_password[:chunk_size] + upper_case_in_password[:chunk_size] + digits_in_password[:chunk_size] + special_chars_in_password[:chunk_size]
        random.shuffle(final_password)

        return ''.join(final_password)

    @classmethod
    def encrypt(self, key, data):
        return Fernet.key(key).encrypt(data)

    @classmethod
    def decrypt(self, key, data):
        return Fernet.key(key).decrypt(data)




# BSD 3-Clause License
# Copyright (c) 2024, mac
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

import os
import logging
import sqlalchemy as db
from datetime import datetime


class CrypterDb(object):
    def __init__(self, db_name, db_path):
        self.db_name = db_name
        self.db_file = f"{os.path.join(db_path, db_name)}.db"
        self.engine = None
        self.connection = None
        self.meta = None

    def __enter__(self):
        self.engine = db.create_engine(f"sqlite:///{self.db_file}")
        self.meta = db.MetaData()
        self.connection = self.engine.connect()
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        if self.connection:
            self.connection.commit()
            self.connection.close()
        self.connection = None
        self.engine = None
        self.meta = None

    def is_db_present(self):
        return os.path.exists(self.db_file)

    def setup(self):
        secret = db.Table("secret", self.meta,
                          db.Column("Id", db.Integer(), nullable=False, primary_key=True, autoincrement=True),
                          db.Column("key", db.String(255), nullable=False),
                          db.Column("created_at", db.DATETIME, nullable=True, default=datetime.now()),
                          db.Column("modified_at", db.DATETIME, nullable=True, default=datetime.now()),
                          db.Column("is_deleted", db.Boolean, nullable=True, default=False),
                          sqlite_autoincrement=True
        )
        records = db.Table("records", self.meta,
                    db.Column("Id", db.Integer(), nullable=False, primary_key=True, autoincrement=True),
                    db.Column("key", db.String(255), nullable=False),
                    db.Column("name", db.String(255), nullable=False),
                    db.Column("password", db.String(255), nullable=False),
                    db.Column("created_at", db.DATETIME, nullable=True, default=datetime.now()),
                    db.Column("modified_at", db.DATETIME, nullable=True, default=datetime.now()),
                    db.Column("is_deleted", db.Boolean, nullable=True, default=False),
                    sqlite_autoincrement=True
        )
        self.meta.create_all(self.engine)

    def execute(self, query):
        result = self.connection.execute(query)
        return result

    def insert(self, table_name, values, return_columns=None):
        if not self.is_db_present():
            raise Exception("No database found. Please run the 'crypter init' command to setup.")
        table = db.Table(table_name, self.meta, autoload_with=self.engine)
        query = table.insert().values(**values)
        if return_columns:
            args = [table.columns[col] for col in return_columns]
            query = query.returning(*args)
        return self.execute(query).fetchall()

    def update(self):
        pass

    def get(self, table_name, key_names, return_columns=None):
        if not self.is_db_present():
            raise Exception("No database found. Please run the 'crypter init' command to setup.")
        table = db.Table(table_name, self.meta, autoload_with=self.engine)
        query = table.select()
        if key_names:
            query = query.where(table.columns.key.in_(key_names))
        if return_columns:
            args = [table.columns[col] for col in return_columns]
            query = query.with_only_columns(*args)
        return self.execute(query).fetchall()

    def delete(self, table_name, key_names, return_columns=None):
        if not self.is_db_present():
            raise Exception("No database found. Please run the 'crypter init' command to setup.")
        table = db.Table(table_name, self.meta, autoload_with=self.engine)
        result = list()
        query = table.delete().where(table.columns.key.in_(key_names))
        if return_columns:
            args = [table.columns[col] for col in return_columns]
            query = query.returning(*args)
        return self.execute(query).fetchall()

    def drop(self, table_name):
        if not self.is_db_present():
            raise Exception("No database found. Please run the 'crypter init' command to setup.")
        table = db.Table(table_name, self.meta, autoload_with=self.engine)
        table.drop(self.engine)
        